//+------------------------------------------------------------------+
//|                                                      fmtprtl.mqh |
//|                                      Copyright  2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//|        MQL's OOP notes: Print data using overloads and templates |
//|                        https://www.mql5.com/en/blogs/post/680570 |
//+------------------------------------------------------------------+

class OutputStream
{
  private:
    int digits;
    char delimiter;
    int timemode;
    string format;
    string line;
    
    void createFormat()
    {
      format = "%." + (string)digits + "f";
    }
    
    string trimmedLine()
    {
      if(delimiter != 0)
      {
        int n = StringLen(line);
        if(n > 0)
        {
          if(StringGetCharacter(line, n - 1) == delimiter)
          {
            return StringSubstr(line, 0, n - 1);
          }
        }
      }
      return line;
    }
   
  public:
    OutputStream(): digits(_Digits), timemode(TIME_DATE|TIME_MINUTES)
    {
      createFormat();
    }
    OutputStream(int p): digits(p), timemode(TIME_DATE|TIME_MINUTES)
    {
      createFormat();
    }
    OutputStream(int p, char d): digits(p), delimiter(d), timemode(TIME_DATE|TIME_MINUTES)
    {
      createFormat();
    }
    OutputStream(int p, char d, int t): digits(p), delimiter(d), timemode(t)
    {
      createFormat();
    }

    template<typename T>
    OutputStream *operator<<(const T v)
    {
      if(typename(v) == "int" || typename(v) == "uint"
      || typename(v) == "long" || typename(v) == "ulong"
      || typename(v) == "short" || typename(v) == "ushort")
      {
        if(typename(v) == "ushort" && v == '\n')
        {
          Print(trimmedLine());
          line = NULL;
          return GetPointer(this);
        }
        else
        {
          line += IntegerToString((int)v);
        }
      }
      else
      if(typename(v) == "double" || typename(v) == "float")
      {
        if(v == EMPTY_VALUE) line += "<EMPTY_VALUE>";
        else line += StringFormat(format, v);
      }
      else
      if(typename(v) == "datetime")
      {
        line += TimeToString((datetime)v, timemode);
      }
      else
      if(typename(v) == "color")
      {
        line += (string)v;
      }
      else
      if(typename(v) == "char" || typename(v) == "uchar")
      {
        if(v == '\n')
        {
          Print(trimmedLine());
          line = NULL;
          return GetPointer(this);
        }
        else
        {
          line += CharToString((char)v);
        }
      }
      else
      {
        line += (string)v;
      }
      if(delimiter != 0)
      {
        line += CharToString(delimiter);
      }
      return GetPointer(this);
    }
    
    template<typename T>
    OutputStream *operator<<(const T &v)
    {
      line += typename(v) + StringFormat("%X", GetPointer(v));
      if(delimiter != 0)
      {
        line += CharToString(delimiter);
      }
      return GetPointer(this);
    }

    template<typename T>
    OutputStream *operator<<(const T *v)
    {
      line += typename(v) + StringFormat("%X", v);
      if(delimiter != 0)
      {
        line += CharToString(delimiter);
      }
      return GetPointer(this);
    }

};

  
template<typename T> string EN(T enum_value)    { return(EnumToString(enum_value)); }

#define EOL (char)'\n'

// USAGE
/*

#property strict

#include <fmtprntl.mqh>

OutputStream out(5, ',', TIME_DATE | TIME_MINUTES | TIME_SECONDS);

void OnStart()
{
  enum days
  {
    one,
    two,
    three
  };

  days day = two;
  bool x = true;
  datetime dt = TimeCurrent();
  color clr = Red;

  out << M_PI << "test" << x << EN(day) << dt << clr << EOL;

}
*/
